"""
CURP Suite
Jacob Sánchez Pérez
Copyright (C) 2021
"""

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

__all__ = [
    "__title__",
    "__summary__",
    "__uri__",
    "__version__",
    "__author__",
    "__email__",
    "__license__",
    "__copyright__",
]

__title__ = "CURPSuite"
__summary__ = "Análisis y validación de la CURP Mexicana"
__uri__ = "https://github.com/jacobszpz/CURPSuite"

__version__ = "2.3.0"

__author__ = "Jacob Sánchez"
__email__ = "jacobszpz@protonmail.com"

__license__ = "GNU General Public License v2"
__copyright__ = f"2021, {__author__}"



#!/usr/bin/env python3

from datetime import date
from unidecode import unidecode
from dataclasses import dataclass
from curp import CURP
from curp.chars import CURPChar
from curp.altisonantes import altisonantes


class FeaturedWord:
    """Similar a :class:`WordFeatures` pero las posiciones de las características
    son especificadas al crear el objeto.

    :param word: Palabra
    :param char: Posición del primer carácter.
    :param vowel: Posición de la primera vocal.
    :param consonant: Posición de la primera consonante.

    Si alguna de las posiciones es -1 la característica asignada será "X"
    """
    def __init__(self, word="", char=0, vowel=1, consonant=2):
        self._word = word
        w = normalise_word(word)
        assert len(w) > max(vowel, consonant)

        self._char = w[char] if w else "X"
        self._vowel = w[vowel] if vowel != -1 else "X"
        self._consonant = w[consonant] if consonant != -1 else "X"

    @property
    def word(self) -> str:
        return self._word

    @property
    def char(self) -> str:
        return self._char

    @property
    def vowel(self) -> str:
        return self._vowel

    @property
    def consonant(self) -> str:
        return self._consonant

    def loosely_not_equal(self, other: 'FeaturedWord') -> bool:
        """Compara aproximadamente si no tiene las mismas características que otra palabra."""
        diff = (self.char != other.char or
                self.consonant != other.consonant)
        return diff

    def __ne__(self, other):
        diff = (self.char != other.char or
                self.vowel != other.vowel or
                self.consonant != other.consonant)
        return diff

    def __repr__(self):
        return f"<FeaturedWord [{self.word}]>"

@dataclass(frozen=True)
class CURPSkeleton:
    """Holds data that belongs to a CURP."""
    curp: str
    name: str
    first_surname: str
    second_surname: str
    features: list[FeaturedWord]
    birth_date: date
    sex: int
    region: dict[str, str]

    @property
    def full_name(self) -> str:
        return f"{self.name} {self.first_surname} {self.second_surname}"


def fix_verification(curp: str) -> str:
    """Corregir dígito de verificación de una CURP."""
    cs = CURP._verification_sum(curp)
    d = CURP._sum_to_verify_digit(cs)
    return f"{curp[:-1]}{d}"

def normalise_word(word: str) -> str:
    """Procesar palabra para eliminar peculiaridades y volverla mayúscula."""
    return unidecode(word.upper().replace("Ñ", "X"))

def build_curp(name: FeaturedWord = None, first_surname: FeaturedWord = None,
               second_surname: FeaturedWord = None, date: date = date(2000, 1, 1),
               sex: str = 'M', region: str = 'DF', fix_digit: bool = True) -> str:
    """Ensambla una CURP a partir de parámetros."""
    n, f, s = name, first_surname, second_surname
    hc = '0' if date.year <= 1999 else 'A'
    curp = f"{f.char}{f.vowel}{s.char}{n.char}"

    # Si la CURP forma palabra inconveniente, censurar
    censored = f"{curp[0]}X{curp[2:]}"
    if censored in altisonantes:
        if curp[1] in altisonantes[censored]:
            curp = censored

    curp += f"{date:%y%m%d}{sex}{region}"
    curp += f"{f.consonant}{s.consonant}{n.consonant}{hc}V"

    if fix_digit:
        curp = fix_verification(curp)

    return curp

def insert_in_word(word: str, c: str, i: int):
    """Reemplaza cierta sección de una string con otra string."""
    return f"{word[:i]}{c}{word[i + len(c):]}"

def change_curp(curp: str, c: str = None, i: int = None, chars: str = None, consonants: str = None,
                date: str = None, sex: str = None, region: str = None):
    """Permite el reemplazo de un elemento de una CURP."""
    if c is not None and i is not None:
        curp = insert_in_word(curp, c, i)

    if chars is not None:
        curp = insert_in_word(curp, chars, CURPChar.SURNAME_A_CHAR)

    if consonants is not None:
        curp = insert_in_word(curp, consonants, CURPChar.SURNAME_A_CONSONANT)

    if date is not None:
        curp = insert_in_word(curp, date, CURPChar.YEAR_0)

    if sex is not None:
        curp = insert_in_word(curp, sex, CURPChar.SEX)

    if region is not None:
        curp = insert_in_word(curp, region, CURPChar.REGION_0)

    # Si es posible, arreglar dígito de verificación
    if curp.isalnum():
        curp = fix_verification(curp)

    return curp



#!/usr/bin/env python3

"""
CURP Suite
~~~~~~~~~~

CURP Suite es una librería para el análisis y validación de la CURP,
o Clave Única de Registro de Población.

Uso básico:
    >>> from curp import CURP
    >>> c = CURP("SABC560626MDFLRN01")
    >>> c.fecha_nacimiento
    datetime.date(1956, 6, 26)
    >>> c.sexo
    2
    >>> c.entidad_nacimiento
    'Ciudad de México'
    >>> c.entidad_nacimiento_iso
    'MX-CMX'

La documentación completa se encuentra en <https://curp.readthedocs.io>.

:copyright: (c) 2021, Jacob Sánchez Pérez.
:license: GPL v2.0, ver LICENSE para más detalles.
"""

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor,
# Boston, MA  02110-1301, USA.
# Also available at https://www.gnu.org/licenses/old-licenses/gpl-2.0.html

from .__about__ import __title__, __summary__, __uri__, __version__
from .__about__ import __author__, __email__, __license__
from .__about__ import __copyright__
from .curp import CURP
from .exceptions import (CURPValueError, CURPLengthError,
                         CURPVerificationError, CURPNameError,
                         CURPFirstSurnameError, CURPSecondSurnameError,
                         CURPFullNameError, CURPDateError, CURPSexError, CURPRegionError)

__all__ = ['__title__', '__summary__', '__uri__', '__version__',
           '__author__', '__email__', '__license__', '__copyright__',
           'CURP', 'CURPValueError', 'CURPLengthError', 'CURPVerificationError',
           'CURPNameError', 'CURPFirstSurnameError', 'CURPSecondSurnameError',
           'CURPFullNameError', 'CURPDateError', 'CURPSexError', 'CURPRegionError']



# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor,
# Boston, MA  02110-1301, USA.

# Copyright (C) 2021, Jacob Sánchez Pérez

from enum import IntEnum


class CURPChar(IntEnum):
    """Guarda las posiciones de los datos de la CURP"""
    # Primer Apellido
    SURNAME_A_CHAR = 0
    SURNAME_A_VOWEL = 1
    SURNAME_A_CONSONANT = 13

    # Segundo Apellido
    SURNAME_B_CHAR = 2
    SURNAME_B_CONSONANT = 14

    # Nombre de Pila
    NAME_CHAR = 3
    NAME_CONSONANT = 15

    # Fecha de Nacimiento
    YEAR_0 = 4
    YEAR_1 = 5
    MONTH_0 = 6
    MONTH_1 = 7
    DAY_0 = 8
    DAY_1 = 9

    # Sexo y Entidad Federativa de Nacimiento
    SEX = 10
    REGION_0 = 11
    REGION_1 = 12

    # Homoclave
    HOMONYMY = 16
    VERIFICATION = 17



#!/usr/bin/env python3

import unittest
from unidecode import unidecode


class TestUnidecode(unittest.TestCase):
    """Probar el método unidecode para remover acentos y diéresis de palabras."""

    def test_accents_removal(self):
        """Probar la remoción de acentos/diéresis de vocales."""
        vowels = {'A': ('Á', 'Ä'), 'E': ('É', 'Ë'), 'I': ('Í', 'Ī'),
                  'O': ('Ó', 'Ö'), 'U': ('Ú', 'Ü')}
        for k, v in vowels.items():
            for c in v:
                self.assertEqual(k, unidecode(c))


if __name__ == '__main__':
    unittest.main()



# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor,
# Boston, MA  02110-1301, USA.

# Copyright (C) 2021, Jacob Sánchez Pérez

from unidecode import unidecode
from functools import partialmethod


class WordFeatures:
    """
    Extrae ciertos caracteres de una palabra:

    - Primer carácter
    - Primera vocal interna
    - Primera consonante interna

    Si no se encuentra alguna de las anteriores,
    una "X" tomará su lugar

    Cualquier "Ñ" será reemplazada con una "X"

    :param str word: Palabra a analizar
    :param tuple[str] ignored_words: Tupla de palabras que no se usarán
    para el análisis si no es absolutamente necesario
    :param tuple[str] special_chars: Caracteres que sirven como separador de
    palabras
    """

    _vowels = "AEIOU"
    _consonants = "BCDFGHJKLMNÑPQRSTVWXYZ"

    def __init__(self, word: str, ignored_words: tuple[str, ...] = (),
                 special_chars: tuple[str, ...] = ()):
        self._char = "X"

        # Reemplazar Ñ's
        word = unidecode(word.upper().replace("Ñ", "X"))

        for c in special_chars:
            word = word.replace(c, " ")

        # Remover preposiciones, conjunciones, etc.
        pieces = word.split()

        if pieces:
            # Preservar última palabra
            pieces = [w for w in pieces[:-1]
                      if w not in ignored_words] + [pieces[-1]]

            # Usar primera palabra del apellido
            word = pieces[0]
            self._char = word[0]

        self._vowel = self._find_vowel(word[1:])
        self._consonant = self._find_consonant(word[1:])

    @staticmethod
    def _find_char(charset: str, word: str, pos: int = 0) -> str:
        """Encuentra cualquier carácter de un set regex en una palabra."""
        # Más rápido que regex para este caso
        for char in word[pos:]:
            if char in charset:
                return char
        return "X"

    _find_vowel = partialmethod(_find_char, _vowels)
    _find_consonant = partialmethod(_find_char, _consonants)

    @property
    def char(self) -> str:
        """Primer carácter."""
        return self._char

    @property
    def vowel(self) -> str:
        """Primera vocal interna."""
        return self._vowel

    @property
    def consonant(self) -> str:
        """Primera consonante interna."""
        return self._consonant



"""
Entidades de la República Mexicana y su código ISO
"""

from typing import Optional, TypedDict


class RegionData(TypedDict):
    name: str
    iso: Optional[str]


estados: dict[str, RegionData] = {
  "AS": {
    "name": "Aguascalientes",
    "iso": "MX-AGU"
  },
  "BC": {
    "name": "Baja California",
    "iso": "MX-BCN"
  },
  "BS": {
    "name": "Baja California Sur",
    "iso": "MX-BCS"
  },
  "CC": {
    "name": "Campeche",
    "iso": "MX-CAM"
  },
  "CL": {
    "name": "Coahuila de Zaragoza",
    "iso": "MX-COA"
  },
  "CM": {
    "name": "Colima",
    "iso": "MX-COL"
  },
  "CS": {
    "name": "Chiapas",
    "iso": "MX-CHP"
  },
  "CH": {
    "name": "Chihuahua",
    "iso": "MX-CHH"
  },
  "DF": {
    "name": "Ciudad de México",
    "iso": "MX-CMX"
  },
  "DG": {
    "name": "Durango",
    "iso": "MX-DUR"
  },
  "GT": {
    "name": "Guanajuato",
    "iso": "MX-GUA"
  },
  "GR": {
    "name": "Guerrero",
    "iso": "MX-GRO"
  },
  "HG": {
    "name": "Hidalgo",
    "iso": "MX-HID"
  },
  "JC": {
    "name": "Jalisco",
    "iso": "MX-JAL"
  },
  "MC": {
    "name": "México",
    "iso": "MX-MEX"
  },
  "MN": {
    "name": "Michoacán de Ocampo",
    "iso": "MX-MIC"
  },
  "MS": {
    "name": "Morelos",
    "iso": "MX-MOR"
  },
  "NT": {
    "name": "Nayarit",
    "iso": "MX-NAY"
  },
  "NL": {
    "name": "Nuevo León",
    "iso": "MX-NLE"
  },
  "OC": {
    "name": "Oaxaca",
    "iso": "MX-OAX"
  },
  "PL": {
    "name": "Puebla",
    "iso": "MX-PUE"
  },
  "QT": {
    "name": "Querétaro",
    "iso": "MX-QUE"
  },
  "QR": {
    "name": "Quintana Roo",
    "iso": "MX-ROO"
  },
  "SP": {
    "name": "San Luis Potosí",
    "iso": "MX-SLP"
  },
  "SL": {
    "name": "Sinaloa",
    "iso": "MX-SIN"
  },
  "SR": {
    "name": "Sonora",
    "iso": "MX-SON"
  },
  "TC": {
    "name": "Tabasco",
    "iso": "MX-TAB"
  },
  "TS": {
    "name": "Tamaulipas",
    "iso": "MX-TAM"
  },
  "TL": {
    "name": "Tlaxcala",
    "iso": "MX-TLA"
  },
  "VZ": {
    "name": "Veracruz de Ignacio de la Llave",
    "iso": "MX-VER"
  },
  "YN": {
    "name": "Yucatán",
    "iso": "MX-YUC"
  },
  "ZS": {
    "name": "Zacatecas",
    "iso": "MX-ZAC"
  },
  "NE": {
    "name": "Extranjero",
    "iso": None
  }
}



#!/usr/bin/env python3

"""CURP Suite Tests"""

# Copyright (C) 2021, Jacob Sánchez Pérez

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import unittest
from hypothesis import given, assume, example
from hypothesis import strategies as st
from .utils import CURPSkeleton, FeaturedWord, change_curp, insert_in_word
from .strategies import WordStrats, CURPStrats, ASCIIStrats

import json
import string
from typing import Union
from datetime import date
from unidecode import unidecode
from curp.curp import WordFeatures
from curp.chars import CURPChar
from curp import CURP, estados, altisonantes
from curp import (CURPValueError, CURPLengthError,
                  CURPVerificationError, CURPNameError,
                  CURPFirstSurnameError, CURPSecondSurnameError,
                  CURPFullNameError, CURPDateError, CURPSexError, CURPRegionError)


class TestCURP(unittest.TestCase):
    """Pruebas de la clase CURP."""

    _common_names = (
        'MARIA', 'MA', 'MA.', 'JOSE', 'J', 'J.',
        'JOSÉ', 'MARÍA', 'JÖSË', 'MÄRÍÁ'
    )

    _charset = string.digits + string.ascii_uppercase

    # Pruebas que requieren que la construcción del objeto falle

    @given(st.text(max_size=17))
    def test_creation_curp_too_short(self, curp: str) -> None:
        """Probar que una CURP muy corta genere error."""
        with self.assertRaises(CURPLengthError):
            CURP(curp)

    @given(st.text(min_size=19))
    def test_creation_curp_too_long(self, curp: str) -> None:
        """Probar que una CURP muy larga genere error."""
        with self.assertRaises(CURPLengthError):
            CURP(curp)

    @given(CURPStrats.curps())
    def test_creation_verification_digit_check(self, sk: CURPSkeleton) -> None:
        """Probar que sólo un dígito de verificación sea válido para cada CURP."""
        d = int(sk.curp[-1])
        for i in range(10):
            if i != d:
                with self.assertRaises(CURPVerificationError):
                    CURP(f"{sk.curp[:-1]}{i}")
            else:
                CURP(sk.curp)

    @given(CURPStrats.curps(), ASCIIStrats.characters(lowercase=False, uppercase=False))
    def test_creation_surname_a_char_error(self, sk: CURPSkeleton, c: str) -> None:
        """Prueba caracteres de nombre/apellidos incorrectos en la CURP."""
        curp = change_curp(sk.curp, c, CURPChar.SURNAME_A_CHAR)

        with self.assertRaises(CURPValueError) as cm:
            CURP(curp)

        self.assertNotIsInstance(cm.exception, CURPVerificationError)

    @given(CURPStrats.curps(), CURPStrats.consonants())
    def test_creation_surname_a_vowel_error(self, sk: CURPSkeleton, c: str) -> None:
        """Prueba caracteres de nombre/apellidos incorrectos en la CURP."""
        assume(c != 'X')
        curp = change_curp(sk.curp, c, CURPChar.SURNAME_A_VOWEL)

        with self.assertRaises(CURPValueError) as cm:
            CURP(curp)

        self.assertNotIsInstance(cm.exception, CURPVerificationError)

    @given(CURPStrats.curps(), CURPStrats.vowels())
    def test_creation_surname_a_consonant_error(self, sk: CURPSkeleton, c: str) -> None:
        """Prueba caracteres de nombre/apellidos incorrectos en la CURP."""
        assume(c != 'X')
        curp = change_curp(sk.curp, c, CURPChar.SURNAME_A_CONSONANT)

        with self.assertRaises(CURPValueError) as cm:
            CURP(curp)

        self.assertNotIsInstance(cm.exception, CURPVerificationError)

    @given(CURPStrats.curps(), ASCIIStrats.characters(lowercase=False, uppercase=False))
    def test_creation_surname_b_char_error(self, sk: CURPSkeleton, c: str) -> None:
        """Prueba caracteres de nombre/apellidos incorrectos en la CURP."""
        curp = change_curp(sk.curp, c, CURPChar.SURNAME_B_CHAR)

        with self.assertRaises(CURPValueError) as cm:
            CURP(curp)

        self.assertNotIsInstance(cm.exception, CURPVerificationError)

    @given(CURPStrats.curps(), CURPStrats.vowels())
    def test_creation_surname_b_consonant_error(self, sk: CURPSkeleton, c: str) -> None:
        """Prueba caracteres de nombre/apellidos incorrectos en la CURP."""
        assume(c != 'X')
        curp = change_curp(sk.curp, c, CURPChar.SURNAME_B_CONSONANT)

        with self.assertRaises(CURPValueError) as cm:
            CURP(curp)

        self.assertNotIsInstance(cm.exception, CURPVerificationError)

    @given(CURPStrats.curps(), ASCIIStrats.characters(lowercase=False, uppercase=False))
    def test_creation_name_char_error(self, sk: CURPSkeleton, c: str) -> None:
        """Prueba caracteres de nombre/apellidos incorrectos en la CURP."""
        curp = change_curp(sk.curp, c, CURPChar.NAME_CHAR)

        with self.assertRaises(CURPValueError) as cm:
            CURP(curp)

        self.assertNotIsInstance(cm.exception, CURPVerificationError)

    @given(CURPStrats.curps(), CURPStrats.vowels())
    def test_creation_name_consonant_error(self, sk: CURPSkeleton, c: str) -> None:
        """Prueba caracteres de nombre/apellidos incorrectos en la CURP."""
        assume(c != 'X')
        curp = change_curp(sk.curp, c, CURPChar.NAME_CONSONANT)

        with self.assertRaises(CURPValueError) as cm:
            CURP(curp)

        self.assertNotIsInstance(cm.exception, CURPVerificationError)

    @given(CURPStrats.curps(), WordStrats.words())
    def test_creation_name_error(self, sk: CURPSkeleton, fake_name: FeaturedWord) -> None:
        """Prueba la creación de una CURP con un nombre de pila incorrecto."""
        # Asumir que el nombre falso no tenga la misma letra inicial
        # y primera consonante interna que el nombre real
        assume(fake_name.loosely_not_equal(sk.features[0]))

        with self.assertRaises(CURPNameError):
            c = CURP(sk.curp, nombre=fake_name.word)

    @given(CURPStrats.curps(), WordStrats.words())
    def test_creation_first_surname_error(self, sk: CURPSkeleton, fake_name: FeaturedWord) -> None:
        """Prueba la creación de una CURP con un primer apellido incorrecto."""
        # Asumir que el nombre falso no tenga
        # las mismas caracteristicas que el nombre real
        assume(fake_name != sk.features[1])

        with self.assertRaises(CURPFirstSurnameError):
            c = CURP(sk.curp, primer_apellido=fake_name.word)

    @given(CURPStrats.curps(), WordStrats.words())
    def test_creation_second_surname_error(self, sk: CURPSkeleton, fake_name: FeaturedWord) -> None:
        """Prueba la creación de una CURP con un primer apellido incorrecto."""
        # Asumir que el nombre falso no tenga la misma letra inicial
        # y primera consonante interna que el nombre real
        assume(fake_name.loosely_not_equal(sk.features[2]))

        with self.assertRaises(CURPSecondSurnameError):
            c = CURP(sk.curp, segundo_apellido=fake_name.word)

    @given(CURPStrats.curps(), st.tuples(WordStrats.words(min_size=1), WordStrats.words(), WordStrats.words()))
    def test_creation_full_name_error(self, sk: CURPSkeleton, fake_names_tuple: tuple[FeaturedWord]) -> None:
        """Prueba la creación de una CURP con un nombre completo incorrecto."""
        assume(sk.first_surname != '')
        assume(sk.second_surname != '')

        fake_names = sorted(fake_names_tuple, reverse=True, key=lambda obj: obj.word)
        # print([n.word for n in fake_names])
        # print(sk.full_name)
        names_different = (fake_names[0].loosely_not_equal(sk.features[0]) or
                           fake_names[1] != sk.features[1] or
                           fake_names[2].loosely_not_equal(sk.features[2]))

        assume(names_different)
        with self.assertRaises(CURPFullNameError):
            c = CURP(sk.curp, nombre_completo=' '.join([name.word for name in fake_names]))

    @given(CURPStrats.curps(), ASCIIStrats.text(min_size=6, max_size=6, lowercase=False))
    def test_creation_date_error(self, sk: CURPSkeleton, d: str)-> None:
        """Prueba la creacion de una CURP con una fecha con caracteres no numericos. """
        assume(any(c not in string.digits for c in d))

        curp = change_curp(sk.curp, date=d)

        with self.assertRaises(CURPValueError) as cm:
            CURP(curp)

        self.assertNotIsInstance(cm.exception, CURPDateError)
        self.assertNotIsInstance(cm.exception, CURPVerificationError)

    @given(CURPStrats.curps(), st.integers(0, 99), st.integers(0, 99), st.integers(0, 99))
    def test_creation_date_error_nonexistent_dates(self, sk: CURPSkeleton, y: int, m: int, d: int) -> None:
        """Probar que fechas incorrectas provoquen un error."""
        date_is_valid = False
        century = '19' if sk.curp[-2].isdigit() else '20'

        try:
            date(int(f"{century}{y:02}"), m, d)
        except ValueError:
            pass
        else:
            date_is_valid = True

        assume(not date_is_valid)

        fake_date = f"{y:02}{m:02}{d:02}"
        curp = change_curp(sk.curp, date=fake_date)

        with self.assertRaises(CURPDateError):
            CURP(curp)

    @given(CURPStrats.curps(), ASCIIStrats.characters(lowercase=False))
    def test_creation_sex_error(self, sk: CURPSkeleton, s: str) -> None:
        """Probar que códigos incorrectos de sexo provoquen un error."""
        # Solo probar códigos alfanuméricos incorrectos
        assume(s != 'H' and s != 'M')

        curp = change_curp(sk.curp, sex=s)

        with self.assertRaises(CURPSexError):
            CURP(curp)

    @given(CURPStrats.curps(), ASCIIStrats.text(min_size=2, max_size=2, lowercase=False))
    def test_creation_region_error(self, sk: CURPSkeleton, r: str) -> None:
        """Probar que códigos incorrectos de región provoquen un error."""
        # Solo probar códigos alfanuméricos incorrectos
        assume(r not in estados.estados)

        curp = change_curp(sk.curp, region=r)

        with self.assertRaises(CURPRegionError):
            CURP(curp)

    @given(st.text(min_size=18, max_size=18))
    def test_creation_invalid_characters(self, curp: str):
        """Probar que caracteres que no son validos en la CURP provoquen un error. """
        assume(any(c not in self._charset for c in curp))

        with self.assertRaises(CURPValueError) as cm:
            CURP(curp)

        self.assertNotIsInstance(cm.exception, CURPVerificationError)
        self.assertNotIsInstance(cm.exception, CURPLengthError)

    # Pruebas que requieren que el objeto sea construido exitosamente

    @given(CURPStrats.curps())
    def test_curp_property(self, sk: CURPSkeleton) -> None:
        """Prueba que la propiedad CURP sea la correcta."""
        c = CURP(sk.curp)
        self.assertEqual(c.curp, sk.curp)

    @given(CURPStrats.curps())
    def test_default_names_are_null(self, sk: CURPSkeleton) -> None:
        """Prueba que los nombres por defecto sean None."""
        c = CURP(sk.curp)
        self.assertIsNone(c.nombre)
        self.assertIsNone(c.primer_apellido)
        self.assertIsNone(c.segundo_apellido)

    @given(CURPStrats.curps())
    def test_name_property(self, sk: CURPSkeleton) -> None:
        """Prueba la creación de una CURP con un nombre de pila."""
        c = CURP(sk.curp, nombre=sk.name)
        self.assertEqual(c.nombre, sk.name.upper())

    @given(CURPStrats.curps())
    def test_first_surname_property(self, sk: CURPSkeleton) -> None:
        """Prueba la creación de una CURP con primer apellido."""
        c = CURP(sk.curp, primer_apellido=sk.first_surname)
        self.assertEqual(c.primer_apellido, sk.first_surname.upper())

    @given(CURPStrats.curps())
    def test_second_surname_property(self, sk: CURPSkeleton) -> None:
        """Prueba la creación de una CURP con segundo apellido."""
        c = CURP(sk.curp, segundo_apellido=sk.second_surname)
        self.assertEqual(c.segundo_apellido, sk.second_surname.upper())

    @given(CURPStrats.curps())
    def test_full_name_properties(self, sk: CURPSkeleton) -> None:
        """Prueba la creación de una CURP con un nombre completo."""
        assume(not self.word_ignored(sk.name))
        assume(not self.word_ignored(sk.first_surname))
        assume(not self.word_ignored(sk.second_surname))

        c = CURP(sk.curp, nombre_completo=sk.full_name)
        self.assertEqual(c.nombre, sk.name.upper())
        self.assertEqual(c.primer_apellido, sk.first_surname.upper())
        self.assertEqual(c.segundo_apellido, sk.second_surname.upper())

    @given(CURPStrats.curps())
    def test_name_and_surnames_properties(self, sk: CURPSkeleton) -> None:
        """Prueba la creación de una CURP con nombres/apellidos en los argumentos."""
        c = CURP(
            sk.curp,
            nombre=sk.name,
            primer_apellido=sk.first_surname,
            segundo_apellido=sk.second_surname
        )

        self.assertEqual(c.nombre, sk.name.upper())
        self.assertEqual(c.primer_apellido, sk.first_surname.upper())
        self.assertEqual(c.segundo_apellido, sk.second_surname.upper())

    @given(CURPStrats.curps())
    def test_birth_date_property(self, sk: CURPSkeleton) -> None:
        """Prueba que la extracción de la fecha de nacimiento de la CURP funcione."""
        c = CURP(sk.curp)
        self.assertEqual(c.fecha_nacimiento, sk.birth_date)

    @given(CURPStrats.curps())
    def test_sex_property(self, sk: CURPSkeleton) -> None:
        """Prueba que la extracción del sexo de la CURP funcione."""
        c = CURP(sk.curp)
        self.assertEqual(c.sexo, sk.sex)

    @given(CURPStrats.curps())
    def test_region_properties(self, sk: CURPSkeleton) -> None:
        """Prueba que la extracción de la entidad federativa de la CURP funcione."""
        c = CURP(sk.curp)
        self.assertEqual(c.entidad, sk.region['name'])
        self.assertEqual(c.entidad_iso, sk.region['iso'])
        self.assertEqual(c.es_extranjero, not bool(sk.region['iso']))

    @given(CURPStrats.curps())
    def test_name_validation(self, sk: CURPSkeleton):
        """Prueba la validación del nombre de pila."""
        c = CURP(sk.curp)
        self.assertTrue(c.nombre_valido(sk.name))

    @given(CURPStrats.curps())
    def test_name_validation_with_compound_common_name(self, sk: CURPSkeleton) -> None:
        """Prueba la comprobación de CURP con un nombre compuesto común."""
        c = CURP(sk.curp)

        # Probar con todos los nombres ignorados
        for n in self._common_names:
            self.assertTrue(c.nombre_valido(f"{n} {sk.name}"))
            self.assertTrue(c.nombre_valido(f"{n.title()} {sk.name}"))
            self.assertTrue(c.nombre_valido(f"{n.lower()} {sk.name}"))

    @given(CURPStrats.curps(), st.text(), CURPStrats.ignored_strings())
    def test_name_validation_with_compound_name(self, sk: CURPSkeleton, n: str, ignored: str) -> None:
        """Prueba la comprobación de CURP con un nombre compuesto."""
        assume(not self.name_ignored(sk.name))
        assume(not self.word_ignored(sk.name))

        c = CURP(sk.curp)
        self.assertTrue(c.nombre_valido(f"{ignored} {sk.name} {n}"))
        self.assertTrue(c.nombre_valido(f"{ignored.title()} {sk.name} {n}"))
        self.assertTrue(c.nombre_valido(f"{ignored.lower()} {sk.name} {n}"))

    @given(CURPStrats.curps())
    def test_first_surname_validation(self, sk: CURPSkeleton):
        """Prueba la validación del primer apellido."""
        c = CURP(sk.curp)
        self.assertTrue(c.primer_apellido_valido(sk.first_surname))

    @given(CURPStrats.curps(), CURPStrats.inconvenient())
    def test_first_surname_validation_inconvenient_word(self, sk: CURPSkeleton, inconvenient: str):
        """Prueba la validación del primer apellido."""
        curp = change_curp(sk.curp, chars=f"{inconvenient[0]}X{inconvenient[2:]}")
        c = CURP(curp)

        first_surname = f"{inconvenient[:2]}{sk.features[1].consonant}"
        self.assertTrue(c.primer_apellido_valido(first_surname))

    @given(CURPStrats.curps(), st.text(), CURPStrats.ignored_strings())
    def test_first_surname_validation_with_compound_surname(self, sk: CURPSkeleton, n: str, ignored: str) -> None:
        """Prueba la comprobación de CURP con un apellido compuesto."""
        assume(sk.first_surname != '')
        assume(not self.word_ignored(sk.first_surname))
        c = CURP(sk.curp)
        self.assertTrue(c.primer_apellido_valido(f"{ignored} {sk.first_surname} {n}"))
        self.assertTrue(c.primer_apellido_valido(f"{ignored.title()} {sk.first_surname} {n}"))
        self.assertTrue(c.primer_apellido_valido(f"{ignored.lower()} {sk.first_surname} {n}"))

    @given(CURPStrats.curps())
    def test_second_surname_validation(self, sk: CURPSkeleton):
        """Prueba la validación del segundo apellido."""
        c = CURP(sk.curp)
        self.assertTrue(c.segundo_apellido_valido(sk.second_surname))

    @given(CURPStrats.curps(), st.text(), CURPStrats.ignored_strings())
    def test_second_surname_validation_with_compound_surname(self, sk: CURPSkeleton, n: str, ignored: str) -> None:
        """Prueba la comprobación de CURP con un apellido compuesto."""
        assume(sk.second_surname != '')
        assume(not self.word_ignored(sk.second_surname))
        c = CURP(sk.curp)
        self.assertTrue(c.segundo_apellido_valido(f"{ignored} {sk.second_surname} {n}"))
        self.assertTrue(c.segundo_apellido_valido(f"{ignored.title()} {sk.second_surname} {n}"))
        self.assertTrue(c.segundo_apellido_valido(f"{ignored.lower()} {sk.second_surname} {n}"))

    @given(CURPStrats.curps())
    def test_full_name_validation(self, sk: CURPSkeleton):
        """Prueba la validación del nombre completo."""
        assume(not self.word_ignored(sk.name))
        assume(not self.word_ignored(sk.first_surname))
        assume(not self.word_ignored(sk.second_surname))

        c = CURP(sk.curp)
        nombre_completo = c.nombre_completo_valido(sk.full_name)
        self.assertTrue(nombre_completo)
        self.assertEqual(len(nombre_completo), 3)

        self.assertEqual(nombre_completo[0], sk.name)
        self.assertEqual(nombre_completo[1], sk.first_surname)
        self.assertEqual(nombre_completo[2], sk.second_surname)

    @given(CURPStrats.curps(), WordStrats.words(), st.one_of(st.none(), CURPStrats.ignored_names()),
           CURPStrats.ignored_strings(), CURPStrats.ignored_strings(), CURPStrats.ignored_strings())
    def test_full_name_validation_with_ignored_words(self, sk: CURPSkeleton, post_name: FeaturedWord, ignored_name: Union[None, str],
                                                     pre_name: str, pre_surname_a: str, pre_surname_b: str):
        """Prueba la validación del nombre completo."""
        assume(not self.word_ignored(sk.name))
        assume(not self.word_ignored(sk.first_surname))
        assume(not self.word_ignored(sk.second_surname))
        assume(sk.features[1] != post_name)
        assume(not self.word_ignored(post_name.word))

        c = CURP(sk.curp)
        first_surname = second_surname = ""

        name_prefix = pre_name

        if ignored_name:
            name_prefix = f"{pre_name} {ignored_name}"

        given_names = f"{name_prefix} {sk.name} {post_name.word}"

        if sk.first_surname:
            first_surname = f"{pre_surname_a} {sk.first_surname}"
        if sk.second_surname:
            second_surname = f"{pre_surname_b} {sk.second_surname}"
        completo = f"{given_names} {first_surname} {second_surname}"

        nombre_completo = c.nombre_completo_valido(completo)
        self.assertTrue(nombre_completo)
        self.assertEqual(len(nombre_completo), 3)

        self.assertEqual(nombre_completo[0], given_names.strip())
        self.assertEqual(nombre_completo[1], first_surname.strip())
        self.assertEqual(nombre_completo[2], second_surname.strip())

    @given(CURPStrats.curps(), WordStrats.words())
    def test_name_validation_false(self, sk: CURPSkeleton, fake_name: FeaturedWord) -> None:
        """Prueba la creación de una CURP con un nombre de pila incorrecto."""
        # Asumir que el nombre falso no tenga la misma letra inicial
        # y primera consonante interna que el nombre real
        assume(fake_name.loosely_not_equal(sk.features[0]))
        c = CURP(sk.curp)
        self.assertFalse(c.nombre_valido(fake_name.word))

    @given(CURPStrats.curps(), WordStrats.words(min_size=1))
    def test_name_validation_with_compound_name_false(self, sk: CURPSkeleton, n: FeaturedWord) -> None:
        """Prueba la comprobación de CURP con un nombre compuesto."""
        assume(sk.features[0].loosely_not_equal(n))
        assume(not self.word_ignored(n.word))
        assume(not self.name_ignored(n.word))

        c = CURP(sk.curp)
        self.assertFalse(c.nombre_valido(f"{n.word} {sk.name}"))

    @given(CURPStrats.curps(), WordStrats.words())
    def test_first_surname_validation_false(self, sk: CURPSkeleton, fake_name: FeaturedWord) -> None:
        """Prueba la creación de una CURP con un primer apellido incorrecto."""
        # Asumir que el nombre falso no tenga la misma letra inicial
        # y primera consonante interna que el nombre real
        assume(fake_name != sk.features[1])
        c = CURP(sk.curp)
        self.assertFalse(c.primer_apellido_valido(fake_name.word))

    @given(CURPStrats.curps(), CURPStrats.inconvenient())
    def test_first_surname_validation_inconvenient_word_false(self, sk: CURPSkeleton, inconvenient: str):
        """Prueba la validación del primer apellido."""
        curp = change_curp(sk.curp, chars=inconvenient)
        with self.assertRaises(CURPValueError):
            c = CURP(curp)

    @given(CURPStrats.curps(), WordStrats.words(min_size=1))
    def test_first_surname_validation_with_compound_surname_false(self, sk: CURPSkeleton, n: FeaturedWord) -> None:
        """Prueba la comprobación de CURP con un apellido compuesto."""
        assume(sk.features[1] != n)
        assume(not self.word_ignored(n.word))
        c = CURP(sk.curp)
        self.assertFalse(c.primer_apellido_valido(f"{n.word} {sk.first_surname}"))

    @given(CURPStrats.curps(), WordStrats.words())
    def test_second_surname_validation_false(self, sk: CURPSkeleton, fake_name: FeaturedWord) -> None:
        """Prueba la creación de una CURP con un primer apellido incorrecto."""
        # Asumir que el nombre falso no tenga la misma letra inicial
        # y primera consonante interna que el nombre real
        assume(fake_name.loosely_not_equal(sk.features[2]))
        c = CURP(sk.curp)
        self.assertFalse(c.segundo_apellido_valido(fake_name.word))

    @given(CURPStrats.curps(), WordStrats.words(min_size=1))
    def test_second_surname_validation_with_compound_surname_false(self, sk: CURPSkeleton, n: FeaturedWord) -> None:
        """Prueba la comprobación de CURP con un apellido compuesto."""
        assume(sk.features[2].loosely_not_equal(n))
        assume(not self.word_ignored(n.word))
        c = CURP(sk.curp)
        self.assertFalse(c.segundo_apellido_valido(f"{n.word} {sk.second_surname}"))

    @given(CURPStrats.curps(), WordStrats.words(min_size=1))
    def test_full_name_validation_wrong_given_name_false(self, sk: CURPSkeleton, n: str):
        """Prueba la validación del nombre completo."""
        assume(not self.word_ignored(sk.name))
        assume(not self.word_ignored(sk.first_surname))
        assume(not self.word_ignored(sk.second_surname))
        assume(not self.word_ignored(n.word))
        assume(not self.name_ignored(n.word))
        assume(sk.features[0].loosely_not_equal(n))

        c = CURP(sk.curp)
        nombre_completo = c.nombre_completo_valido(f"{n.word} {sk.full_name}")
        self.assertFalse(nombre_completo)

    @given(CURPStrats.curps())
    def test_full_name_validation_first_surname_not_empty_false(self, sk: CURPSkeleton):
        """Prueba que la validación del nombre completo falle si el apellido existe y no se encuentra."""
        empty_word = FeaturedWord('X', vowel=-1, consonant=-1)
        assume(sk.features[1] != empty_word)
        assume(not self.word_ignored(sk.name))
        assume(not self.word_ignored(sk.first_surname))

        c = CURP(sk.curp)
        nombre_completo = c.nombre_completo_valido(f"{sk.name}")
        self.assertFalse(nombre_completo)

    @given(CURPStrats.curps())
    def test_full_name_validation_second_surname_not_empty_false(self, sk: CURPSkeleton):
        """Prueba que la validación del nombre completo falle si el apellido existe y no se encuentra."""
        empty_word = FeaturedWord('X', vowel=-1, consonant=-1)
        assume(sk.features[2].loosely_not_equal(empty_word))
        assume(not self.word_ignored(sk.name))
        assume(not self.word_ignored(sk.first_surname))
        assume(not self.word_ignored(sk.second_surname))

        c = CURP(sk.curp)
        nombre_completo = c.nombre_completo_valido(f"{sk.name} {sk.first_surname}")
        self.assertFalse(nombre_completo)

    @given(CURPStrats.curps())
    def test_json_dump(self, sk: CURPSkeleton):
        """Prueba el método json."""
        c = CURP(sk.curp)

        j = json.loads(c.json())
        self.assertEqual(c.curp, j['curp'])
        self.assertEqual(str(c.fecha_nacimiento), j['fecha_nacimiento'])
        self.assertEqual(c.sexo, j['sexo'])
        self.assertEqual(c.entidad, j['entidad_nacimiento']['name'])
        self.assertEqual(c.entidad_iso, j['entidad_nacimiento']['iso'])

        n = any([x in j for x in ('nombre', 'primer_apellido', 'segundo_apellido')])
        self.assertFalse(n)

    @given(CURPStrats.curps())
    def test_json_dump_name_and_surnames(self, sk: CURPSkeleton):
        """Prueba el método json."""
        c = CURP(
            sk.curp,
            nombre=sk.name,
            primer_apellido=sk.first_surname,
            segundo_apellido=sk.second_surname
        )

        j = json.loads(c.json())
        self.assertEqual(c.nombre, j['nombre'])
        self.assertEqual(c.primer_apellido, j['primer_apellido'])
        self.assertEqual(c.segundo_apellido, j['segundo_apellido'])

    # Pruebas de clase

    @given(st.integers(0))
    def test_static_sum_to_digit(self, s: int):
        """Probar método para convertir suma en dígito verificador."""
        d = CURP._sum_to_verify_digit(s)
        self.assertEqual(len(d), 1)
        self.assertTrue(d.isdigit())

        # Alternativa a mod
        m = int(str(s)[-1])
        m = 10 - m if m else m
        self.assertEqual(str(m), d)

    def test_static_verification_sum(self):
        """Probar la comprobación correcta del dígito de verificación
        cuando la CURP contiene sólo un carácter distinto a cero."""
        charset = string.digits + string.ascii_uppercase

        # Probar cada carácter válido de la CURP en cada posición.
        for c in charset:
            b36 = int(c, 36)
            if b36 > 23:
                b36 += 1

            for n in range(2, 19):
                with self.subTest(c=c, n=n):
                    # La suma sólo tendrá un elemento.
                    sm = b36 * n

                    # Crear CURP
                    curp = ("0" * (18 - n)) + c + ("0" * (n - 1))
                    cs = CURP._verification_sum(curp)
                    self.assertEqual(sm, cs)

    # Otras pruebas

    # @given(CURPStrats.curps())
    # def test_example(self, g):
    #     print(g)

    # Utilidades

    @staticmethod
    def word_ignored(word: str):
        return unidecode(word.upper()) in CURP._ignored_words

    @staticmethod
    def name_ignored(name: str):
        return unidecode(name.upper()) in CURP._ignored_names


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3

import unittest


class TestCURPExamples(unittest.TestCase):

    def test_inconvenient_words(self) -> None:
        pass


    # def test_name_splitting(self):
    #     curp_items = [
    #         {'curp': 'POPC990709MGTSRL02',
    #          'name': 'CLAUDIA LEONOR POSADA PEREZ',
    #          'result': {'nombre': 'CLAUDIA LEONOR', 'apellido': 'POSADA', 'apellido_m': 'PEREZ'}
    #         },
    #         {'curp': 'MAGE981117MMNCRS05',
    #          'name': 'ESTEFANIA DE LOS DOLORES MACIAS GARCIA',
    #          'result': {'nombre': 'ESTEFANIA DE LOS DOLORES', 'apellido': 'MACIAS', 'apellido_m': 'GARCIA'}
    #         },
    #         {'curp': 'MAPS991116MOCRZN07',
    #          'name': 'SANDRA DEL CARMEN MARTINEZ DE LA PAZ',
    #          'result': {'nombre': 'SANDRA DEL CARMEN', 'apellido': 'MARTINEZ', 'apellido_m': 'DE LA PAZ'}
    #         },
    #         {'curp': 'TAXA990915MNEMXM06',
    #          'name': 'AMBER NICOLE TAMAYO',
    #          'result': {'nombre': 'AMBER NICOLE', 'apellido': 'TAMAYO', 'apellido_m': ''}
    #         },
    #         {'curp': 'MXME991209MGRRSS07',
    #          'name': 'ESMERALDA MARTINEZ MASTACHE',
    #          'result': {'nombre': 'ESMERALDA', 'apellido': 'MARTINEZ', 'apellido_m': 'MASTACHE'}
    #         },
    #         {'curp': 'MACD990727MMMCRRN0',
    #          'name': 'DANIELA IVETTE MARTINEZ CRUZ',
    #          'result': False
    #         }
    #     ]
    #
    #     for i in curp_items:
    #         with self.subTest(i=i):
    #             r = CURPValidator.validate(i['curp'], i['name'])
    #             self.assertEqual(r, i['result'])



#!/usr/bin/env python3

import argparse
from .curp import CURP, CURPValueError


def main() -> int:
    parser = argparse.ArgumentParser(
        'curp',
        description='Extraer datos de una CURP y validarla.')

    parser.add_argument('curp', help='la curp a analizar')
    parser.add_argument('-n', '--nombre', help='nombre de pila para validar la CURP')
    parser.add_argument('-p', '--primer-apellido', help='primer apellido para validar la CURP')
    parser.add_argument('-s', '--segundo-apellido', help='segundo apellido para validar la CURP')
    parser.add_argument('-c', '--nombre-completo', help='nombre completo para validar la CURP')

    args = parser.parse_args()

    if args.curp is not None:
        try:
            c = CURP(
                args.curp,
                nombre=args.nombre,
                primer_apellido=args.primer_apellido,
                segundo_apellido=args.segundo_apellido,
                nombre_completo=args.nombre_completo
            )
        except CURPValueError as e:
            print(e)
        else:
            print(c.json())
    return 0


if __name__ == '__main__':
    exit(main())



#!/usr/bin/env python3

import unittest
from hypothesis import given

from curp import CURP
from curp.features import WordFeatures
from .utils import FeaturedWord
from .strategies import WordStrats


class TestWordFeatures(unittest.TestCase):
    """Pruebas de la clase WordFeatures."""

    @given(WordStrats.words())
    def test_word_features_extraction(self, featured_word: FeaturedWord) -> None:
        """Usar generador de palabras para probar la clase :class:`WordFeatures`."""
        wf = WordFeatures(featured_word.word)
        self.assertEqual(featured_word.char, wf.char)
        self.assertEqual(featured_word.vowel, wf.vowel)
        self.assertEqual(featured_word.consonant, wf.consonant)

    def test_word_features_extraction_examples(self) -> None:
        """Probar la extracción correcta de características de palabras."""
        words = [
            # Generales
            ('ALVIRDE', 'A', 'I', 'L'),
            ('JAUREGUI', 'J', 'A', 'R'),
            ('ESTRADA', 'E', 'A', 'S'),

            # Minúscula
            ('anna', 'A', 'A', 'N'),
            ('marcial', 'M', 'A', 'R'),

            # Acentos
            ('Álvaro', 'A', 'A', 'L'),
            ('maría', 'M', 'A', 'R'),

            # Letra inicial Ñ
            ('ÑANDO', 'X', 'A', 'N'),

            # Carácteres especiales (/, -, ., ')
            ('D/AMICO', 'D', 'X', 'X'),
            ("d’Arc", 'D', 'X', 'X'),
            ("B'IVANNY", 'B', 'X', 'X'),
            ('AL-SHAMI', 'A', 'X', 'L'),
            ('H. LUZ', 'H', 'X', 'X'),

            # Diéresis
            ('ARGÜELLO', 'A', 'U', 'R'),
            ('MÄNNIG', 'M', 'A', 'N'),
            ('Gëzim', 'G', 'E', 'Z'),
            ('Ingrïd', 'I', 'I', 'N'),
            ('WÖRNE', 'W', 'O', 'R'),
            ('DÖNICKE', 'D', 'O', 'N'),

            # Primera palabra de apellido compuesto
            ('RIVA PALACIO', 'R', 'I', 'V'),
            ('CABEZA DE VACA', 'C', 'A', 'B'),
            ('PONCE DE LEON', 'P', 'O', 'N'),
            ('MONTES DE OCA', 'M', 'O', 'N'),

            # Nombre compuesto con preposición
            # (DA, DAS, DE, DEL, DER, DI, DIE, DD,
            # EL, LA, LOS, LAS, LE, LES, MAC, MC,
            # VAN, VON, Y)
            ('da Silva', 'S', 'I', 'L'),
            ('DAS', 'D', 'A', 'S'),
            ('DE ANDA', 'A', 'A', 'N'),
            ('De Niro', 'N', 'I', 'R'),
            ('de Haas', 'H', 'A', 'S'),
            ('DEL CASTILLO', 'C', 'A', 'S'),
            ('de La Motte', 'M', 'O', 'T'),
            ('DE LOS SANTOS', 'S', 'A', 'N'),
            ('DE LAS FUENTES', 'F', 'U', 'N'),
            ('DE LAS CUEVAS', 'C', 'U', 'V'),
            ('der Ältere', 'A', 'E', 'L'),
            ('Di María', 'M', 'A', 'R'),
            ('EL CARMEN', 'C', 'A', 'R'),
            ('EL GERGI', 'G', 'E', 'R'),
            ('EL ATI', 'A', 'I', 'T'),
            ('La Rose', 'R', 'O', 'S'),
            ('LE LOREC', 'L', 'O', 'R'),
            ('LE SOTO', 'S', 'O', 'T'),
            ('LE CLERCQ', 'C', 'E', 'L'),
            ('Mac Laren', 'L', 'A', 'R'),
            ('MC GREGOR', 'G', 'E', 'R'),
            ('van Leeuwen', 'L', 'E', 'W'),
            ('van der Waals', 'W', 'A', 'L'),
            ('VAN DER PLAS', 'P', 'A', 'L'),
            ('van Beethoven', 'B', 'E', 'T'),
            ('van der Rohe', 'R', 'O', 'H'),
            ('Von Doom', 'D', 'O', 'M'),
            ('von der Vogelweide', 'V', 'O', 'G'),
            ('Y LOPEZ', 'L', 'O', 'P'),
            ('Y VARGAS', 'V', 'A', 'R'),
            ('Y MONTIEL', 'M', 'O', 'N'),

            # Preposición sola
            ('DA', 'D', 'A', 'X'),
            ('DAS', 'D', 'A', 'S'),
            ('DE', 'D', 'E', 'X'),
            ('DEL', 'D', 'E', 'L'),
            ('DER', 'D', 'E', 'R'),
            ('DI', 'D', 'I', 'X'),
            ('DIE', 'D', 'I', 'X'),
            ('DD', 'D', 'X', 'D'),
            ('EL', 'E', 'X', 'L'),
            ('LA', 'L', 'A', 'X'),
            ('LOS', 'L', 'O', 'S'),
            ('LAS', 'L', 'A', 'S'),
            ('LE', 'L', 'E', 'X'),
            ('LES', 'L', 'E', 'S'),
            ('MAC', 'M', 'A', 'C'),
            ('MC', 'M', 'X', 'C'),
            ('VAN', 'V', 'A', 'N'),
            ('VON', 'V', 'O', 'N'),
            ('Y', 'Y', 'X', 'X'),

            # No son preposición
            ('DU SOLIER', 'D', 'U', 'X'),
            ('van den Berg', 'D', 'E', 'N'),

            # Sin vocal interna
            ('ICH', 'I', 'X', 'C'),
            ('ALF', 'A', 'X', 'L'),
            ('ANN', 'A', 'X', 'N'),
            ('AXL', 'A', 'X', 'X'),
            ('IVY', 'I', 'X', 'V'),

            # Consonante interna es Ñ
            ('OÑATE', 'O', 'A', 'X'),
            ('PEÑA', 'P', 'E', 'X'),
            ('IÑIGO', 'I', 'I', 'X'),
            ('NUÑO', 'N', 'U', 'X'),
            ('MUÑOZ', 'M', 'U', 'X'),

            # Sin consonante interna
            ('PO', 'P', 'O', 'X'),
            ('LE', 'L', 'E', 'X'),
            ('HO', 'H', 'O', 'X'),
            ('YU', 'Y', 'U', 'X'),
            ('WU', 'W', 'U', 'X'),
            ('LI', 'L', 'I', 'X'),
            ('LEE', 'L', 'E', 'X'),
            ('DEE', 'D', 'E', 'X'),

            # Sin apellido
            ('', 'X', 'X', 'X')
        ]

        for w in words:
            with self.subTest(w=w[0]):
                ft = WordFeatures(
                    w[0],
                    CURP._ignored_words,
                    CURP._special_chars
                )
                self.assertEqual(w[1], ft.char)
                self.assertEqual(w[2], ft.vowel)
                self.assertEqual(w[3], ft.consonant)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3

import string
from datetime import date
from unidecode import unidecode
from curp import CURP
from curp import estados, altisonantes
from .utils import normalise_word, build_curp, FeaturedWord, CURPSkeleton
from hypothesis import given
from hypothesis.strategies import (sampled_from, composite, tuples,
                                   integers, dates, lists, booleans)


class ASCIIStrats:
    """Contiene estrategias de :class:`hypothesis` para generar texto ascii."""

    @classmethod
    @composite
    def text(draw, cls, min_size: int = 0, max_size: int = None,
             lowercase: bool = True, uppercase: bool = True, digits: bool = True) -> str:
        """Genera texto ascii."""
        charset = ""

        if digits:
            charset += string.digits
        if lowercase:
            charset += string.ascii_lowercase
        if uppercase:
            charset += string.ascii_uppercase

        t = ''.join(draw(lists(sampled_from(charset), min_size=min_size, max_size=max_size)))
        return t

    @classmethod
    @composite
    def characters(draw, cls, lowercase: bool = True, uppercase: bool = True, digits: bool = True):
        return draw(cls.text(min_size=1, max_size=1, lowercase=lowercase, uppercase=uppercase, digits=digits))


class WordStrats:
    """Contiene estrategias de :class:`hypothesis` para generar palabras."""

    # Charsets
    _consonants = 'BCDFGHJKLMNÑPQRSTVWXYZ'
    _vowels = 'AÁÄEÉËIÍÏOÓÖUÚÜ'

    # Duplicar los charsets ahora en minúscula
    _consonants += _consonants.lower()
    _vowels += _vowels.lower()
    _alphabet = _vowels + _consonants

    @classmethod
    @composite
    def words(draw, cls, min_size: int = 0, max_size: int = None) -> FeaturedWord:
        """Estrategia para generar palabras con características
        (1ra letra, vocal interna y consonante interna).

        :param min_size: Longitud mínima de palabra, por defecto es 0.
        :param max_size: Longitud máxima de palabra.
        :return: Objeto que contiene palabra y sus características.
        :rtype: :class:`FeaturedWord`
        """

        # Empezar por generar una lista de booleanos
        bool_list = draw(lists(booleans(), min_size=min_size, max_size=max_size))
        # Convertir booleanos a vocales/consonantes, creando así una palabra
        word = [draw(cls.vowels()) if x else draw(cls.consonants()) for x in bool_list]

        # Obtener posiciones de primera consonante y vocal internas
        try:
            first_consonant = bool_list[1:].index(False) + 1
        except ValueError:
            first_consonant = -1
        try:
            first_vowel = bool_list[1:].index(True) + 1
        except ValueError:
            first_vowel = -1

        fw = FeaturedWord(
            word=''.join(word), vowel=first_vowel, consonant=first_consonant
        )
        return fw

    @classmethod
    @composite
    def letters(draw, cls) -> str:
        """Letras del alfabeto español."""
        return draw(sampled_from(cls._alphabet))

    @classmethod
    @composite
    def vowels(draw, cls) -> str:
        """Vocales del alfabeto español. Incluye versiones con acentos y diéresis."""
        return draw(sampled_from(cls._vowels))

    @classmethod
    @composite
    def consonants(draw, cls) -> str:
        """Consonantes del alfabeto español."""
        return draw(sampled_from(cls._consonants))


class CURPStrats:
    """Estrategias de :class:`hypothesis` para generar la CURP Mexicana
    y sus atributos."""

    _regions = estados.estados
    _altisonantes = altisonantes.altisonantes
    _sexes = ('H', 'M')
    _consonants = 'BCDFGHJKLMNPQRSTVWXYZ'
    _vowels = 'AEIOUX'

    @classmethod
    @composite
    def curps(draw, cls) -> CURPSkeleton:
        """Estrategia que genera Claves Únicas de Registro de Población.
        :return: Dataclass que contiene la CURP y los datos de la misma."""

        surnames = []

        # Generar nombre y dos apellidos
        w = draw(WordStrats.words(min_size=1))
        given_name = w

        for _ in range(2):
            w = draw(WordStrats.words())
            surnames.append(w)

        # Asegurar que no existan casos en los que
        # el primer apellido este vacío y el segundo no
        surnames.sort(reverse=True, key=lambda obj: obj.word)

        sex = draw(cls.sexes())
        date = draw(cls.birth_dates())
        region = draw(cls.mexican_states())

        curp = build_curp(
            name=given_name,
            first_surname=surnames[0],
            second_surname=surnames[1],
            date=date,
            sex=sex[0],
            region=region[0]
        )

        sk = CURPSkeleton(
            curp=curp,
            name=given_name.word,
            first_surname=surnames[0].word,
            second_surname=surnames[1].word,
            features=[given_name, *surnames],
            birth_date=date,
            sex=sex[1],
            region=region[1]

        )

        return sk

    @classmethod
    @composite
    def birth_dates(draw, cls) -> date:
        """Estrategia que genera objetos :class:`datetime.date` entre
        1900-01-01 y 2099-12-31."""
        return draw(dates(min_value=date(1900, 1, 1), max_value=date(2099, 12, 31)))

    @classmethod
    @composite
    def sexes(draw, cls) -> tuple[str, int]:
        """Estrategia que genera claves de sexo válidas para la CURP.
        :return: Tupla conteniendo la clave de sexo y su valor numérico.
        """
        i = draw(integers(1, len(cls._sexes)))
        return (cls._sexes[i - 1], i)

    @classmethod
    @composite
    def mexican_states(draw, cls) -> tuple[str, dict[str, str]]:
        """Estrategia que genera claves de entidad federativa válidas para la CURP.
        :return: Tupla conteniendo la clave de entidad federativa y
            un diccionario de datos de la entidad.
        """
        a = sampled_from(sorted(cls._regions.keys()))
        k = draw(a)
        v = cls._regions[k]
        return (k, v)

    @classmethod
    @composite
    def ignored_words(draw, cls) -> str:
        """Palabras ignoradas en la CURP."""
        return draw(sampled_from(CURP._ignored_words))

    @classmethod
    @composite
    def ignored_names(draw, cls) -> str:
        """Palabras ignoradas en la CURP."""
        return draw(sampled_from(CURP._ignored_names))

    @classmethod
    @composite
    def ignored_strings(draw, cls, min_size=0, max_size=None) -> str:
        """Cadena de palabras ignoradas."""
        return draw(lists(cls.ignored_words(), min_size=min_size, max_size=max_size).map(' '.join))

    @classmethod
    @composite
    def inconvenient(draw, cls) -> str:
        """Palabras inconvenientes de la CURP."""
        k = draw(sampled_from(sorted(cls._altisonantes.keys())))
        v = draw(sampled_from(cls._altisonantes[k]))
        return f"{k[0]}{v}{k[2:]}"

    @classmethod
    @composite
    def vowels(draw, cls) -> str:
        """Vocales que pueden aparecer en la CURP."""
        return draw(sampled_from(cls._vowels))

    @classmethod
    @composite
    def consonants(draw, cls) -> str:
        """Consonantes que pueden aparecer en la CURP."""
        return draw(sampled_from(cls._consonants))




"""
Clases de excepciones de CURP Suite
"""


# Clase base
class CURPValueError(ValueError):
    """Indica que la validación de la CURP presentó errores."""


class CURPLengthError(CURPValueError):
    """La CURP no tiene la longitud correcta de 18 caracteres."""


class CURPVerificationError(CURPValueError):
    """El dígito verificador de la CURP no es el calculado."""


class CURPNameError(CURPValueError):
    """El nombre provisto no corresponde a la CURP."""


class CURPFirstSurnameError(CURPValueError):
    """El primer apellido provisto no corresponde a la CURP."""


class CURPSecondSurnameError(CURPValueError):
    """El segundo apellido provisto no corresponde a la CURP."""


class CURPFullNameError(CURPValueError):
    """El nombre completo provisto no corresponde a la CURP."""


class CURPDateError(CURPValueError):
    """La fecha indicada en la CURP es incorrecta."""


class CURPSexError(CURPValueError):
    """El sexo indicado en la CURP no es válido."""


class CURPRegionError(CURPValueError):
    """La entidad federativa indicada en la CURP no es válida."""



"""
CURP Suite
Análisis y validación de la CURP mexicana.
"""

# Copyright (C) 2021, Jacob Sánchez Pérez

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor,
# Boston, MA  02110-1301, USA.


import json
import string
from datetime import date
from enum import Enum, auto
from unidecode import unidecode
from typing import Optional, Union, Literal

from .chars import CURPChar
from .features import WordFeatures
from . import altisonantes, estados
from .exceptions import (CURPValueError, CURPLengthError,
                         CURPVerificationError, CURPNameError,
                         CURPFirstSurnameError, CURPSecondSurnameError,
                         CURPFullNameError, CURPDateError, CURPSexError,
                         CURPRegionError)

__all__ = ["CURP"]


class CURP():
    """
    Realiza extracción de datos y validación de una CURP
    (Clave Única de Registro de Población).

    Uso::

        >>> from curp import CURP
        >>> c = CURP("SABC560626MDFLRN01")
        >>> c
        <CURP [SABC]>
    """
    _LENGTH = 18
    # str.index solo regresa la primera coincidencia
    _CHARSET = f"{string.digits}{string.ascii_uppercase[:14]}"
    _CHARSET += f"N{string.ascii_uppercase[14:]}"

    _ignored_words = ('DA', 'DAS', 'DE', 'DEL', 'DER', 'DI', 'DIE', 'DD',
                      'EL', 'LA', 'LOS', 'LAS', 'LE', 'LES', 'MAC', 'MC',
                      'VAN', 'VON', 'Y')
    _special_chars = ('/', '-', '.', "'", '’')
    _ignored_names = ('MARIA', 'MA', 'MA.', 'JOSE', 'J', 'J.')

    # Sexos
    _sexes: dict[str, Literal[1, 2]] = {'H': 1, 'M': 2}

    # Abreviaciones de los estados en la CURP
    _regions = estados.estados

    # Palabras inconvenientes de la CURP
    _inconvenient = altisonantes.altisonantes

    # Para detectar si una CURP es inválida debido a estar sin censura,
    # crear lista de palabras sin censura
    _inc_uncensored = [f"{k[0]}{vowel}{k[2:]}"
                       for k, vowels in _inconvenient.items() for vowel in vowels]

    class _NameParseState(Enum):
        """Indica el progreso al comparar un nombre completo con una CURP."""
        NONE = auto()
        GIVEN_NAMES = auto()
        FIRST_SURNAME = auto()
        SECOND_SURNAME = auto()

    def __init__(self, curp: str, nombre: str = None,
                 primer_apellido: str = None, segundo_apellido: str = None,
                 nombre_completo: str = None):
        """
        Construye una CURP.

        Si sólo se proporciona un nombre completo, se dividirá de acuerdo a la CURP.
        Si se proporciona el nombre por partes, se usarán en lugar del nombre completo.
        Sólo se validarán las partes que se proporcionen.

        :param str curp: Una CURP de 18 caracteres.
        :param str nombre: Nombre de pila de la persona.
        :param str primer_apellido: Primer apellido (paterno) de la persona.
        :param str segundo_apellido: Segundo apellido (materno) de la persona.
        :param str nombre_completo: Nombre completo de la persona.
        """
        self._curp = curp

        if len(curp) != self._LENGTH:
            raise CURPLengthError("La CURP no tiene el tamaño correcto")

        if not self._validate_verify():
            raise CURPVerificationError("El dígito verificador no "
                                        "coincide con la CURP")

        # Fecha de nacimiento
        self._parse_birth_date()
        # Sexo
        self._parse_sex()
        # Estado de la república
        self._parse_region()

        # Nombre
        self._name = self._first_surname = self._second_surname = None

        # Validar caracteres restantes
        if not self._validate_name_chars():
            raise CURPValueError('Los caracteres del nombre/apellidos contienen errores')

        if nombre is not None:
            if self.nombre_valido(nombre):
                self._name = nombre.upper()
            else:
                raise CURPNameError('El nombre de pila no coincide con la CURP')

        if primer_apellido is not None:
            if self.primer_apellido_valido(primer_apellido):
                self._first_surname = primer_apellido.upper()
            else:
                raise CURPFirstSurnameError('El primer apellido no coincide con la CURP')

        if segundo_apellido is not None:
            if self.segundo_apellido_valido(segundo_apellido):
                self._second_surname = segundo_apellido.upper()
            else:
                raise CURPSecondSurnameError('El segundo apellido no coincide con la CURP')

        no_pieces = nombre is primer_apellido is segundo_apellido is None

        if no_pieces and nombre_completo is not None:
            names = self.nombre_completo_valido(nombre_completo)

            if names:
                capital_names = [n.upper() for n in names]
                self._name, self._first_surname, self._second_surname = capital_names
            else:
                raise CURPFullNameError('El nombre completo no parece coincidir con la CURP')

    def __repr__(self) -> str:
        return f"<CURP [{self.curp[:CURPChar.NAME_CHAR + 1]}]>"

    def nombre_valido(self, name: str) -> bool:
        """Verifica que una CURP sea válida para cierto nombre de pila.

        :param str name: Nombre de pila para validar.

        .. todo::

            Averiguar si la regla de los nombres comunes aplica aún si hay palabras
            ignoradas antes de ellos.
        """
        # Remover primer nombre si este es muy común
        pieces = name.upper().split()

        # TODO: Tal vez sea necesario omitir palabras ignoradas (De, Del, Etc.)
        if len(pieces) > 1 and unidecode(pieces[0]) in self._ignored_names:
            pieces.pop(0)

        wf = WordFeatures(' '.join(pieces),
                          self._ignored_words,
                          self._special_chars)

        valid = ((self.curp[CURPChar.NAME_CHAR] == wf.char)
                 and (self.curp[CURPChar.NAME_CONSONANT] == wf.consonant))
        return valid

    def primer_apellido_valido(self, primer_apellido: str) -> bool:
        """Verifica que una CURP sea válida para cierto primer apellido.

        :param str primer_apellido: Primer apellido (usualmente paterno) para validar.
        """
        curp_start = self.curp[:CURPChar.NAME_CHAR + 1]

        wf = WordFeatures(primer_apellido, self._ignored_words, self._special_chars)

        valid = ((self.curp[CURPChar.SURNAME_A_CHAR] == wf.char)
                 and (self.curp[CURPChar.SURNAME_A_CONSONANT] == wf.consonant))

        if valid:
            valid = self.curp[CURPChar.SURNAME_A_VOWEL] == wf.vowel

            # Buscar principio de la CURP en la lista de palabras inconvenientes
            if curp_start in self._inconvenient:
                # Usar las vocales reales para determinar si la CURP
                # corresponde al apellido
                for vowel in self._inconvenient[curp_start]:
                    valid = valid or (vowel == wf.vowel)

        return valid

    def segundo_apellido_valido(self, segundo_apellido: str) -> bool:
        """Verifica que una CURP sea válida para cierto segundo apellido.

        :param str segundo_apellido: Segundo apellido (usualmente materno) para validar.
        """
        wf = WordFeatures(segundo_apellido,
                          self._ignored_words,
                          self._special_chars)

        valid = ((self.curp[CURPChar.SURNAME_B_CHAR] == wf.char)
                 and (self.curp[CURPChar.SURNAME_B_CONSONANT] == wf.consonant))
        return valid

    def nombre_completo_valido(self, nombre_completo: str) -> Union[tuple[str, ...], Literal[False]]:
        """Utiliza un nombre completo para validar la CURP.

        :param str nombre_completo: Nombre completo para validar.
        :return: Una tupla con el nombre por partes, o False si el nombre no corresponde.
        :rtype: tuple or False
        """
        NameParseState = self._NameParseState

        names: dict[CURP._NameParseState, list[str]] = {k: [] for k in list(NameParseState)}
        state_gen = (s for s in NameParseState)
        state = next(state_gen)
        validation = {NameParseState.NONE: self.nombre_valido,
                      NameParseState.GIVEN_NAMES: self.primer_apellido_valido,
                      NameParseState.FIRST_SURNAME: self.segundo_apellido_valido,
                      NameParseState.SECOND_SURNAME: lambda n: False}

        # Guarda preposiciones y otras palabras ignoradas
        # hasta encontrar siguiente palabra normal
        ignored_buffer: list[str] = []

        for word in nombre_completo.split():
            if unidecode(word.upper()) not in self._ignored_words:
                if validation[state](word):
                    state = next(state_gen)
                elif state is NameParseState.NONE:
                    if unidecode(word.upper()) not in self._ignored_names:
                        return False

                # Agregar palabras ignoradas guardadas a la parte actual
                if len(ignored_buffer):
                    names[state].extend(ignored_buffer)
                    ignored_buffer.clear()
                names[state].append(word)

            else:
                ignored_buffer.append(word)

        names[state].extend(ignored_buffer)

        valid = state is NameParseState.SECOND_SURNAME

        if state is NameParseState.FIRST_SURNAME:
            valid = self.segundo_apellido_vacio
        elif state is NameParseState.GIVEN_NAMES:
            valid = self.primer_apellido_vacio

        if valid:
            names_list = [names[n] for n in NameParseState]
            names_list[0:2] = [names_list[0] + names_list[1]]
            names_tuple = tuple([' '.join(x) for x in names_list])
            return names_tuple

        return False

    def _validate_name_chars(self) -> bool:
        """Valida que los caracteres correspondientes al nombre y apellidos
        estén dentro del espacio correcto."""

        consonants = WordFeatures._consonants
        vowels = WordFeatures._vowels + 'X'

        name_chars = ((CURPChar.NAME_CHAR, CURPChar.NAME_CONSONANT),
                      (CURPChar.SURNAME_A_CHAR, CURPChar.SURNAME_A_CONSONANT),
                      (CURPChar.SURNAME_B_CHAR, CURPChar.SURNAME_B_CONSONANT))

        valid = self.curp[CURPChar.SURNAME_A_VOWEL] in vowels

        for char, consonant in name_chars:
            valid = valid and (self.curp[char] in string.ascii_uppercase and
                               self.curp[consonant] in consonants)

        if self.curp[:CURPChar.NAME_CHAR + 1] in self._inc_uncensored:
            valid = False

        return valid

    def _parse_birth_date(self) -> date:
        """Obtiene la fecha de nacimiento de la CURP.

        :raises ValueError: La fecha de nacimiento no pudo ser construída
        :return: La fecha de nacimiento indicada en la CURP
        :rtype: datetime.date
        """
        # Homonímia
        # [0-9] para personas nacidas hasta el 1999
        # [A-Z] para personas nacidas desde el 2000
        homonymy = self.curp[CURPChar.HOMONYMY]
        before_2k = homonymy in string.digits

        # Día y mes de nacimiento
        try:
            day = int(self.curp[CURPChar.DAY_0] + self.curp[CURPChar.DAY_1])
            month = int(self.curp[CURPChar.MONTH_0] + self.curp[CURPChar.MONTH_1])
            year = int(self.curp[CURPChar.YEAR_0] + self.curp[CURPChar.YEAR_1])
        except ValueError:
            raise CURPValueError(
                "La fecha de nacimiento contiene caracteres no numéricos")

        # Año y siglo actual
        current_year = date.today().year
        century = int(current_year / 100)

        # Asume que cualquier año mayor al actual
        # es en realidad del siglo pasado
        if year > (current_year % 100):
            century -= 1

        # Aunque no necesariamente cierto,
        # es probablemente la mejor opción
        if before_2k:
            century = 19
        elif century == 19:
            century = 20

        year += (century * 100)

        # Regresar con error si la fecha es incorrecta
        try:
            birth_date = date(year, month, day)
        except ValueError:
            raise CURPDateError("La fecha de nacimiento es incorrecta")

        self._birth_date = birth_date
        return birth_date

    def _parse_sex(self) -> Literal[1, 2]:
        """Obtiene el sexo de la CURP.

        :raises ValueError: El sexo en la CURP es incorrecto.
        :return: El sexo de acuerdo a ISO/IEC 5218.
        :rtype: int
        """
        curp_sex = self.curp[CURPChar.SEX]
        sex: Literal[0, 1, 2] = self._sexes.get(curp_sex, 0)

        if not sex:
            raise CURPSexError("El sexo de la CURP no es válido")

        self._sex = sex
        return sex

    def _parse_region(self) -> estados.RegionData:
        """Obtiene la entidad federativa de nacimiento de la CURP.

        :raises ValueError: La CURP contiene un código de entidad incorrecto.
        :return: El nombre y código ISO 3166-2 de la entidad federativa
        de nacimiento.
        :rtype: dict
        """
        curp_region = self.curp[CURPChar.REGION_0] + self.curp[CURPChar.REGION_1]
        region = self._regions.get(curp_region, None)

        if not region:
            raise CURPRegionError("La entidad de nacimiento es incorrecta")

        self._birth_place = region
        return region

    def _validate_verify(self) -> bool:
        """Usa el último carácter de la CURP para verificar la misma
        de acuerdo al algoritmo oficial.
        """
        # Código de verificación
        verify = self.curp[CURPChar.VERIFICATION]
        b37_sum = self._verification_sum(self.curp)
        # Hacer las operaciones finales
        return self._sum_to_verify_digit(b37_sum) == verify

    @classmethod
    def _verification_sum(cls, curp: str) -> int:
        """Suma de verificación de la CURP."""
        try:
            r = sum([i * cls._CHARSET.index(x) for i, x in enumerate(curp[-2::-1], 2)])
            # Asegurarse de que el dígito de verificación sea válido
            cls._CHARSET.index(curp[-1])
        except ValueError:
            raise CURPValueError("La CURP contiene caracteres no válidos.")
        else:
            return r

    @staticmethod
    def _sum_to_verify_digit(sm: int) -> str:
        d = sm % 10
        d = 10 - d if d else d
        return str(d)

    def json(self) -> str:
        """Objeto JSON conteniendo los datos extraídos de la CURP."""
        json_data = {
            'curp': self.curp,
            'sexo': self.sexo,
            'fecha_nacimiento': str(self.fecha_nacimiento),
            'entidad_nacimiento': self._birth_place
        }

        if self._name is not None:
            json_data['nombre'] = self._name

        if self._first_surname is not None:
            json_data['primer_apellido'] = self._first_surname

        if self._second_surname is not None:
            json_data['segundo_apellido'] = self._second_surname
        return json.dumps(json_data, ensure_ascii=False)

    # Properties

    @property
    def curp(self) -> str:
        """CURP con la que se construyó el objeto."""
        return self._curp

    @property
    def nombre(self) -> Optional[str]:
        """Nombre con el que se construyó el objeto."""
        return self._name

    @property
    def primer_apellido(self) -> Optional[str]:
        """Primer apellido con el que se construyó el objeto."""
        return self._first_surname

    @property
    def segundo_apellido(self) -> Optional[str]:
        """Segundo apellido con el que se construyó el objeto."""
        return self._second_surname

    @property
    def fecha_nacimiento(self) -> date:
        """Fecha de nacimiento extraída de la CURP."""
        return self._birth_date

    @property
    def sexo(self) -> Literal[1, 2]:
        """Sexo extraído de la CURP."""
        return self._sex

    @property
    def entidad(self) -> str:
        """Entidad federativa de nacimiento de la CURP."""
        return self._birth_place['name']

    @property
    def entidad_iso(self) -> Optional[str]:
        """Código ISO de la entidad federativa de nacimiento de la CURP."""
        return self._birth_place['iso']

    @property
    def es_extranjero(self) -> bool:
        """True si la CURP pertenece a alguien nacido en el extranjero."""
        return not self._birth_place['iso']

    @property
    def primer_apellido_vacio(self) -> bool:
        """True si la CURP puede corresponder a un primer apellido vacio."""
        return self.primer_apellido_valido('')

    @property
    def segundo_apellido_vacio(self) -> bool:
        """True si la CURP puede corresponder a un segundo apellido vacio."""
        return self.segundo_apellido_valido('')



#!/usr/bin/env python3
from setuptools import setup
from curp.__about__ import (__title__, __version__, __summary__, __uri__, __author__, __email__, __license__)

setup(
    name=__title__,
    version=__version__,
    description=__summary__,
    url=__uri__,
    author=__author__,
    author_email=__email__,
    license=__license__)
